<?php
class GSPMgr
{
    private static $assets = [];

    private static $custom_roles = [];

    private static $_loaded = false;
    private static $_custom_roles_loaded = false;

    public const TYPE_JS = 'js';
    public const TYPE_CSS = 'css';

    public const ROLE_HOME = 'home';
    public const ROLE_ALL = 'all';
    public const ROLE_ARCHIVE = 'archive';
    public const ROLE_CATEGORY = 'category';
    public const ROLE_SINGLE = 'single';
    public const ROLE_PAGE = 'page';
    public const ROLE_WOOCOMMERCE = 'woocommerce';
    public const ROLE_SEARCH = 'search';

    public const POSITION_HEADER = 'header';
    public const POSITION_FOOTER = 'footer';


    public static function reset()
    {
        self::$_loaded = false;
        self::$assets = [];
    }

    public static function deleteRole($role)
    {
        $role = sanitize_text_field($role);
        delete_option("gspmgr-assets-{$role}");
        delete_option("gspmgr-assets-{$role}-exists");

        if (isset(self::$custom_roles[$role]) !== false) {
            unset(self::$custom_roles[$role]);
            self::saveCustomRoles();
        }

        return $role;
    }

    public static function saveCustomRole($role_name, $url_list)
    {
        $crrole = isset(self::$custom_roles[$role_name]) ? self::$custom_roles[$role_name] : new GspRole($role_name, false);
        $crrole->urls = $url_list;
        self::$custom_roles[$crrole->getKey()] = $crrole;
        self::saveCustomRoles();
        return $crrole;
    }

    public static function saveCustomRoles()
    {
        update_option("gsp-custom-roles", self::$custom_roles);
    }

    public static function getAllAssets()
    {
        return self::$assets;
    }

    public static function load($role)
    {
        if (self::$_loaded) return;
        self::$assets = get_option("gspmgr-assets-{$role}", []);
        self::loadCustomRoles();
        self::$_loaded = true;
    }

    public static function loadCustomRoles()
    {
        if (self::$_custom_roles_loaded) return;

        self::$custom_roles = get_option("gsp-custom-roles", []);
        self::$_custom_roles_loaded = true;
    }

    public static function saveAssets($role)
    {
        update_option("gspmgr-assets-{$role}", self::$assets);
        update_option("gspmgr-assets-{$role}-exists", true);
        return true;
    }

    public static function roleExists($role)
    {
        return get_option("gspmgr-assets-{$role}-exists", false);
    }

    /**
     * @return GspAsset
     */
    public static function getAsset(string $url, string $type, string $role)
    {
        $asset = new GspAsset($url, $type, $role);
        $key = $asset->getKey();
        if (isset(self::$assets[$key])) {
            return self::$assets[$key];
        } else {
            return null;
        }
    }

    public static function setAsset($url, $type, $role, $disabled = null, $position = null, $async = null,$defer = null)
    {
        $asset = self::getAsset($url, $type, $role);
        if (!$asset) $asset = new GspAsset($url, $type, $role);
        if ($disabled !== null) {
            $asset->disabled = $disabled;
        }
        if ($position) {
            $asset->position = $position;
        }
        if ($async !== null) {
            $asset->async = $async;
        }
        if ($defer !== null) {
            $asset->defer = $defer;
        }

        self::$assets[$asset->getKey()] = $asset;
    }

    public static function removeAsset(GspAsset $asset)
    {
        unset(self::$assets[$asset->getKey()]);
    }


    private static function readUrl($url)
    {
        $agent = "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1)";

        // Execute
        $content = wp_remote_get($url, [
            'user-agent' => $agent,
            'timeout' => 15,
            'sslverify' => false
        ]);

        return isset($content['body']) ? $content['body'] : null;
    }
    public static function parseUrlAssets($url, $sort = true)
    {

        $content = self::readUrl($url);

        if (!$content) {
            return ['success' => false, 'msg' => 'Could not read url ' . $url];
        }

        if (strpos($content, 'cloudflare-static/rocket-loader.min.js') !== false) {
            return ['success' => false, 'msg' => 'Please disable cloudflare rocket-loader in order to use this plugin, you can enable it once you finish optimization'];
        }


        //scripts
        $result = ['scripts' => [], 'styles' => []];;
        $links = [];
        $scripts = [];

        if (preg_match_all('/<link.+?rel=["\']stylesheet["\'].+?href=["\'](.+?)["\']/', $content, $matches)) {
            $links = array_merge($links, $matches[1]);
        }
        if (preg_match_all('/<link.+?href=["\'](.+?)["\'].+?rel=["\']stylesheet["\']/', $content, $matches)) {
            $links = array_merge($links, $matches[1]);
        }

        if (preg_match_all('/<script.+?src=["\'](.+?)["\']/', $content, $matches)) {
            $scripts = array_merge($scripts, $matches[1]);
        }

        if ($sort) {
            asort($scripts);
            asort($links);
        }

        foreach ($scripts as $script) {
            if (strpos($script, 'gspmgr/js/gspmgr.js') !== false || strpos($script, 'gspmgr/js/gspmgr.min.js') !== false) {
                continue;
            }
            $result['scripts'][] = gsp_cleanUrl($script);
        }

        foreach ($links as $link) {
            if (strpos($script, 'gspmgr/css/gspmgr.css') !== false || strpos($script, 'gspmgr/css/gspmgr.min.css') !== false) {
                continue;
            }
            $result['styles'][] = gsp_cleanUrl($link);
        }


        $result['success'] = true;

        return $result;
    }


    /**
     * @return GspRole[]
     */
    public static function getSysRoles()
    {
        $sys_roles = [self::ROLE_ALL, self::ROLE_HOME, self::ROLE_CATEGORY, self::ROLE_ARCHIVE, self::ROLE_PAGE, self::ROLE_SINGLE, self::ROLE_WOOCOMMERCE, self::ROLE_SEARCH];

        $roles = array_map(function ($rolename) {
            return new GspRole($rolename, true);;
        }, $sys_roles);

        return $roles;
    }

    /**
     * @return GspRole[]
     */
    public static function getRoleList()
    {
        self::loadCustomRoles();
        $roles = array_merge(self::getSysRoles(), self::$custom_roles);

        return $roles;
    }


    /**
     * @return GspRole
     */
    public static function findRole($role_key)
    {
        self::loadCustomRoles();
        $roles = self::getRoleList();
        foreach ($roles as $role) {
            if ($role->getKey() == $role_key) return $role;
        }
        return null;
    }


    /**
     * @return GspRole
     */
    public static function getCurrentUrlAppliedRole()
    {
        //check custom roles first
        self::loadCustomRoles();
        foreach (self::$custom_roles as $gsprole) {
            if ($gsprole->isMatch() && self::roleExists($gsprole->getKey())) {
                return $gsprole;
            }
        }

        foreach (self::getSysRoles() as $gsprole) {
            if ($gsprole->isMatch() && self::roleExists($gsprole->getKey())) {
                return $gsprole;
            }
        }

        return null;
    }
    /**
     * @return GspRole[]
     */
    public static function getCurrentUrlMatchedRoles()
    {
        $roles = [];
        //check custom roles first
        self::loadCustomRoles();
        foreach (self::$custom_roles as $gsprole) {
            if ($gsprole->isMatch()) {
                $roles[] = $gsprole;
            }
        }

        foreach (self::getSysRoles() as $gsprole) {
            if ($gsprole->isMatch()) {
                $roles[] = $gsprole;
            }
        }

        return $roles;
    }

    public static function filterOutput($output, $url, $softDelete = false)
    {
        $role = self::getCurrentUrlAppliedRole();
        if (!$role) $role = new GspRole(self::ROLE_ALL, true);
        $role_key = $role->getKey();
        self::reset();
        self::load($role_key);

        if (preg_match_all('/<script[^<]+?src=[\'"](.+?)[\'"].*>.*?<\/script>/i', $output, $matchSripts)) {
            foreach ($matchSripts[1] as $script) {
                $clean_script = gsp_cleanUrl($script);
                if ($asset = self::getAsset($clean_script, self::TYPE_JS, $role_key)) {
                    $regex = '/<script.+?src=[\'"]' . preg_quote($script, '/') . '.*[\'"].*?>.*?<\/script>/i';
                    if ($asset->disabled) {
                        $output = preg_replace($regex, '', $output);
                    } else {
                        if (preg_match($regex, $output, $matches)) {
                            $assetcontent = $matches[0];
                            if ($asset->async) {
                                $assetcontent_updated = str_replace('<script ', '<script async ', $assetcontent);
                                $output = str_replace($assetcontent, $assetcontent_updated, $output);
                                $assetcontent = $assetcontent_updated;
                            }

                            if ($asset->defer) {
                                $assetcontent_updated = str_replace('<script ', '<script defer ', $assetcontent);
                                $output = str_replace($assetcontent, $assetcontent_updated, $output);
                                $assetcontent = $assetcontent_updated;
                            }

                            if ($asset->position) {
                                $output = self::moveAsset($output, $assetcontent, $asset->position);
                            }
                        }
                    }
                }
            }
        }


        if (preg_match_all('/<link[^<]+?href=[\'"](.+?)[\'"].*?\/?>/i', $output, $matchLinks)) {
            foreach ($matchLinks[1] as $link) {
                $clean_link = gsp_cleanUrl($link);

                if ($asset = self::getAsset($clean_link, self::TYPE_CSS, $role_key)) {
                    $regex = '/<link[^<]+?href=[\'"]' . preg_quote($link, '/') . '.*?[\'"].*?\/?>/i';

                    if ($asset->disabled) {
                        if ($softDelete) {
                            if (preg_match($regex, $output, $matches)) {
                                $assetcontent = $matches[0];
                                $assetcontent_updated = preg_replace('/rel=".+?"/', 'rel="deleted"', $assetcontent);
                                $assetcontent_updated = preg_replace('/rel=\'.+?\'/', 'rel="deleted"', $assetcontent);

                                $output = str_replace($assetcontent, $assetcontent_updated, $output);
                            }
                        } else {
                            $output = preg_replace($regex, '', $output);
                        }
                    } else {
                        if (preg_match($regex, $output, $matches)) {
                            $assetcontent = $matches[0];
                            $assetcontent_updated = $assetcontent;
                            if ($asset->async || $asset->defer) {
                                if (preg_match('/rel=["\'](.+?)["\']/', $assetcontent, $matches)) {
                                    $assetcontent_updated = str_replace($matches[0], '', $assetcontent_updated);
                                }

                                $assetcontent_updated = str_replace('<link ', "<link rel='preload' as='style' onload='this.onload=null;this.rel=\"stylesheet\"'", $assetcontent_updated);

                                $output = str_replace($assetcontent, $assetcontent_updated, $output);
                                $assetcontent = $assetcontent_updated;
                            }

                            if ($asset->position) {
                                $output = self::moveAsset($output, $assetcontent, $asset->position);
                            }
                        }
                    }
                }
            }
        }

        return $output;
    }


    private static function moveAsset($html, $assetcontent, $position)
    {
        $html = str_replace($assetcontent, '', $html);
        if ($position == self::POSITION_FOOTER) {
            $ary = explode('</body>', $html);
            $ary[count($ary) - 2] .= $assetcontent;
            $html = implode("\n</body>", $ary);
        } else {
            $ary = explode('</head>', $html);
            $ary[count($ary) - 2] .= $assetcontent;
            $html = implode("\n</head>", $ary);
        }
        return $html;
    }
}



class GspRole
{
    private $name;
    public $urls;
    private $key;
    private $systemRole;

    function toArray()
    {
        return ['name' => $this->name, 'urls' => $this->urls, 'key' => $this->key];
    }

    function __construct(string $name, bool $systemRole)
    {
        $this->name = $name;
        $this->systemRole = $systemRole;
        $this->key = self::makeKey($name);
    }

    function isSystemRole()
    {
        return $this->systemRole;
    }

    public function getName()
    {
        return $this->name;
    }

    function getKey()
    {
        return $this->key;
    }

    public function isMatch()
    {
        if (!$this->isSystemRole()) {
            if ($this->urls && is_array($this->urls)) {
                $currentUrl = gsp_remove_gsp_url_params(gsp_getCurrentUrl());
                $currentUrl = urldecode($currentUrl);
                $currentUrl = gsp_cleanUrl($currentUrl);
                foreach ($this->urls as $url) {
                    $url = gsp_remove_gsp_url_params($url);
                    $url = gsp_cleanUrl($url);

                    $urlreg = preg_quote($url, "/");
                    $urlreg = str_replace("\\*", ".*", $urlreg);
                    if (preg_match("/$urlreg/i", $currentUrl)) {
                        return true;
                    }
                }
            }
        } else {
            //if no custom role matched, let's go with system roles based on key
            switch ($this->getKey()) {
                case GSPMgr::ROLE_HOME:
                    if (((is_home() || is_front_page()))) {
                        return true;
                    };
                    break;

                case GSPMgr::ROLE_WOOCOMMERCE:
                    if (function_exists('is_woocommerce') && is_woocommerce()) {
                        return true;
                    };
                    break;
                case GSPMgr::ROLE_PAGE:
                    if (is_page()) {
                        return true;
                    };
                    break;
                case GSPMgr::ROLE_CATEGORY:
                    if (is_category()) {
                        return true;
                    };
                    break;
                case GSPMgr::ROLE_ARCHIVE:
                    if (is_archive()) {
                        return true;
                    };
                    break;
                case GSPMgr::ROLE_SINGLE:
                    if (is_archive()) {
                        return true;
                    };
                    break;
                case GSPMgr::ROLE_SEARCH:
                    if (is_search()) {
                        return true;
                    };
                    break;
            }
        }
        return false;
    }

    static function makeKey($name)
    {

        $invchars = array(
            ' ', ',', '_', '*', '/', '\\', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '+', '~'
        );

        $mkey = $name;
        foreach ($invchars as $c) {
            $mkey = str_replace($c, '-', $mkey);
        }

        $mkey = str_replace('--', '-', $mkey);
        $mkey = strtolower($mkey);
        $mkey = trim($mkey, " \t\n\r\0\x0B-.");
        $mkey = trim($mkey, " \t\n\r\0\x0B-.");
        $mkey = trim($mkey, " \t\n\r\0\x0B-.");
        return $mkey;
    }
}


class GspAsset
{
    public $url;
    public $type;
    public $role;
    public $disabled = null;
    public $position = null;
    public $async = null;
    public $defer = null;

    function __construct(string $url, string $type, string $role)
    {
        $this->url = $url;
        $this->role = $role;
        $this->type = $type;
    }

    public function getKey()
    {
        return $this->url . '_' . $this->type;
    }
}
